#!/usr/bin/env python
"""

"""
# Author: Harco Kuppens


import sys, re, pprint  # modules from standard lib (python 2.6 and later)
import yaml
import argparse
import tempfile
import shutil


import uppaal2bcg
import uppaal2aut
import model_parser 
import register 

# load libraries from ./lib
import lib.util,lib.check,lib.run,lib.error


# TODO: add other lts types with nice lts2lts script
import learnlib_dot2jtorx_aut,jtorx_aut2cadp_aut
#import lts2lts


# explicit define public api for when used as module
__all__ = [ 'compare' ]
# note: main and argparsing are just helpers for the function give_user_acces_to_dir_using_acl

def enum(**enums):
        return type('Enum', (), enums)

OutputType = enum(JSON=1, BOOLEAN=2, TEXT=3)
ModelType = enum(IOMEALY=1, LTS=2)

def args_parsing(args):
    import os.path
    defOutput=OutputType.JSON
    defModel=ModelType.IOMEALY
    argparser = argparse.ArgumentParser(prog=os.path.basename(args[0]))
    argparser.add_argument('--verbose', action='store_true', help='verbose output. Only visible in text output mode.')
    argparser.add_argument('--keep-tmp-files', action='store_true', help='keep temporary files generated for doing comparison')
    argparser.add_argument('--skip-bothways', action='store_true', help='only compare model1 against model2; skipping model2 against model1; bothways comparing only done for IOMEALY models to find both differing outputs')

    group=argparser.add_mutually_exclusive_group()
    group.add_argument('--output-json', action='store_const', const=OutputType.JSON, dest='output_type', default=defOutput, help='output json (default)')
    group.add_argument('--output-boolean', action='store_const', const=OutputType.BOOLEAN, dest='output_type', default=defOutput, help='output only boolean equal value: "True" or "False"')
    group.add_argument('--output-text', action='store_const', const=OutputType.TEXT, dest='output_type', default=defOutput, help='output text')

    groupModel=argparser.add_mutually_exclusive_group()
    groupModel.add_argument('--lts', action='store_const', const=ModelType.LTS, dest='model_type', default=defModel, help='specify model is a label transition system')
    groupModel.add_argument('--iomealy', action='store_const', const=ModelType.IOMEALY, dest='model_type', default=defModel, help='specify model is an io automata describing a mealy automata (default)')

    argparser.add_argument('modelfile1', help='model file which can be one of : uppaal model,cadp aut, jtorx aut, dot')
    argparser.add_argument('modelfile2', help='model file which can be one of : uppaal model,cadp aut, jtorx aut, dot')
    argparser.add_argument('range_min', nargs="?", default="0", help='minimum value of range used for input params when flattening the model. Only needed for models with data. (default: 0)')
    argparser.add_argument('range_max', nargs="?", default="0", help='maximum value of range used for input params when flattening the model. Only needed for models with data. (default: 0)')

    parse_result=argparser.parse_args(args[1:])

    verbose=parse_result.verbose
    keeptmp=parse_result.keep_tmp_files
    skip_bothways=parse_result.skip_bothways
    
    output_type=parse_result.output_type
    model_type=parse_result.model_type

    if output_type==OutputType.BOOLEAN:
        skip_bothways=True



    # handle input/output file argument
    modelfile1=os.path.abspath(parse_result.modelfile1).replace('\\','/')
    modelfile2=os.path.abspath(parse_result.modelfile2).replace('\\','/')
    # Note: on windows also interprets / as dir sep, and returns path with windows \


    lib.check.check_isreadablefile(modelfile1)
    lib.check.check_isreadablefile(modelfile2)

    range_min=parse_result.range_min
    range_max=parse_result.range_max

    return [modelfile1,modelfile2,range_min,range_max,keeptmp,verbose,skip_bothways,output_type,model_type]


def model2bcg(input,outdir,name,range_min,range_max,keeptmp,verbose,isLTS):
    output_str=""

    isIOMealyInLTS = not isLTS
    # IOMealyInLTS is an io mealy model in LTS file format where both input and output are 
    # on a single label:  label="input/output" 

    bcgfile=outdir + '/' + name + '.bcg'
    if input.endswith('.xml'):
        uppfile=""
        modeltype=model_parser.get_modeltype(input)
        if modeltype == "uppaal":
            uppfile=input
        elif modeltype == "register":
            uppfile=outdir + '/' + name + '.uppaal.xml'
            regmodel=register.from_xml(input)
            regmodel.to_uppaalfile(uppfile)
        else:
            print "error: this xml input model file is not supported. Xml model file must be either an 'uppaal' or 'register' model."
            exit(1)

        output_str+=uppaal2bcg.uppaal2bcg(uppfile,bcgfile,range_min,range_max,keeptmp,verbose)
        return output_str
    if input.endswith('.bcg'):
        shutil.copy(input,bcgfile)
        return "just take copy of input bcg file"
    
# TODO: add other lts types with nice lts2lts script
    if input.endswith('.jtorx.aut'):
        cadp_aut_file=outdir + '/' + name + '.cadp.aut'
        jtorx_aut2cadp_aut.jtorx_aut2cadp_aut(input,cadp_aut_file)
        command=[ 'bcg_io',  cadp_aut_file,bcgfile ]
        if verbose: 
            output_str+="convert jtrox.aut to cadp.aut "  + "\n"
            output_str+="cwd : " +  outdir + "\n"
            output_str+="run command (cadp.aut2bcg) : " +  str(command) + "\n"
        lib.run.run_command(command,cwd=outdir)
    elif input.endswith('.tomte.aut'):     # if not specified assumes tomte : I/O and '_' separator   => changed : now you must specify .tomte.aut
        cadp_aut_file=outdir + '/' + name + '.aut'
        uppaal2aut.tomte_aut2cadp_aut(input,cadp_aut_file)
        command=[ 'bcg_io',  cadp_aut_file,bcgfile ]
        if verbose: 
            output_str+="convert tomte_aut to cadp.aut "  + "\n"
            output_str+="cwd : " +  outdir + "\n"
            output_str+="run command (cadp.aut2bcg) : " +  str(command) + "\n"
        lib.run.run_command(command,cwd=outdir)
    elif input.endswith('.aut') and isLTS: # cadp.aut or lts.aut
        command=[ 'bcg_io',  input,bcgfile ]
        if verbose: 
            output_str+="cwd : " +  outdir + "\n"
            output_str+="run command (for lts: aut2bcg) : " +  str(command) + "\n"
        lib.run.run_command(command,cwd=outdir)
    elif input.endswith('.dot') and isIOMealyInLTS:
        # => split MEALY transitions before compare
        jtorx_aut_file=outdir + '/' + name + '.jtorx.aut'
        learnlib_dot2jtorx_aut.dot2jtorxaut(input,jtorx_aut_file)
        if verbose: 
            output_str+="convert .splitIOMEALY.dot to jtorx.aut" + "\n"
        output_str+=model2bcg(jtorx_aut_file,outdir,name,range_min,range_max,keeptmp,verbose,isLTS=True) # from now on handle as lts in conversion
    elif input.endswith('.dot') and isLTS:
        lts_aut_file=outdir + '/' + name + '.lts.aut'
        learnlib_dot2jtorx_aut.lts_dot2aut(input,lts_aut_file)
        if verbose: 
            output_str+="convert an LTS from  dot to .lts.aut" + "\n"
        output_str+=model2bcg(lts_aut_file,outdir,name,range_min,range_max,keeptmp,verbose,isLTS)
    else:
        print "format for input '" + input + "' not supported"
        sys.exit(1) 
        
    return output_str

def generateModelFiles(modelfile1,modelfile2,range_min,range_max,tmpdir,keeptmp=False,verbose=False,isLTS=False):
    line="\n-----------------------------------------------\n" 
    output_str="flatten first model to a cadp bcg file" 
    output_str+=line
    output_str+=model2bcg(modelfile1,tmpdir,'model1',range_min,range_max,keeptmp,verbose,isLTS)
    output_str+="flatten second model to a cadp bcg file" 
    output_str+=line
    output_str+=model2bcg(modelfile2,tmpdir,'model2',range_min,range_max,keeptmp,verbose,isLTS)
    return output_str.strip()


def compareModels(tmpdir,reverse=False,verbose=False):
    line="\n-----------------------------------------------\n" 
    model1 = 'model1.bcg'
    model2 = 'model2.bcg'
    if reverse:
        (model1,model2)=(model2,model1)
    command = ['time', 'bcg_open', model1 , 'bisimulator', '-diag', '-bfs', '-strong',  model2  ]

    output_str="compare " + model1 + " with " + model2 + " with cadp caesar" 
    output_str+=line
    if verbose:
        output_str+="tmpdir: " + tmpdir + "\n"
        output_str+="cwd : " +  tmpdir + "\n"
        output_str+="run command: " + str(command) + "\n"
    output_str+=lib.run.run_command_with_output(command,cwd=tmpdir)
    return output_str.strip()
   
def getTraceFromText_LTS(s): 
        # get Counter example from text
        start=s.rfind('<start state>')+14
        end=s.rfind('<goal state>')
        s.rfind('<goal state>')
        actions=s[start:end].translate(None,'\x01\x02').strip().split('\n')  #"\x01"
        # relabel labels from cadp format to tomte format
        actions=[label.strip('"').replace(' !','_') for label in actions]
    
        #pprint.pprint(actions)
        actions[-1]=actions[-1].replace('Present in model2.bcg: ','')
        actions[-1]=actions[-1].replace('Present in model1.bcg: ','')
    
        return actions

def getTraceFromText_IOMEALY(s): 
        # get Counter example from text
        start=s.rfind('<start state>')+14
        end=s.rfind('<goal state>')
        s.rfind('<goal state>')
        ioactions=s[start:end].translate(None,'\x01\x02').strip().split('\n')  #"\x01"
        # relabel labels from cadp format to tomte format
        ioactions=[label.strip('"').replace(' !','_') for label in ioactions]
    
        #pprint.pprint(ioactions)
    
        # remove outputs by only allowing even entries
        inputs=ioactions[::2]
        outputs=ioactions[1::2]
  
        #print 'inputs'
        #pprint.pprint(inputs)
        #print 'outputs'
        #pprint.pprint(outputs)

        # differening input/output has extra text 'Present in <modelname>.bcg'
        # where modelname can be eithre model1 or model2 depending on the order
        # we now remove this 
        if len(inputs) > len(outputs)   :
            # differ by input 
            inputs[-1]=inputs[-1].replace('Present in model2.bcg: ','')
            inputs[-1]=inputs[-1].replace('Present in model1.bcg: ','')
        else: 
            # differ by output 
            outputs[-1]=outputs[-1].replace('Present in model2.bcg: ','')
            outputs[-1]=outputs[-1].replace('Present in model1.bcg: ','')

        #pprint.pprint(inputs)
        #pprint.pprint(outputs)
    
        return (inputs,outputs)

def compare_lts(modelfile1,modelfile2,range_min,range_max,keeptmp=False,verbose=False):
    tmpdir=tempfile.mkdtemp(prefix="sut_compare_")

    s=generateModelFiles(modelfile1,modelfile2,range_min,range_max,tmpdir,keeptmp,verbose,isLTS=True)
    output_str=s
    s=compareModels(tmpdir,True,verbose=verbose)
    output_str=output_str+s
    equal=s.endswith('TRUE')

    actionsModel1=None


    if not equal:
        actionsModel1=getTraceFromText_LTS(s) 


    if not keeptmp:
        shutil.rmtree(tmpdir)

    # remove the control chars from the cadp output
    output_str=''.join(c for c in output_str if ord(c) >= 10)

    return {'equal':equal,'actionsModel1':actionsModel1,'fullOutput':output_str}

def compare_IOMEALY(modelfile1,modelfile2,range_min,range_max,keeptmp=False,verbose=False,bothways=True):
    tmpdir=tempfile.mkdtemp(prefix="sut_compare_")

    s=generateModelFiles(modelfile1,modelfile2,range_min,range_max,tmpdir,keeptmp,verbose,isLTS=False)
    output_str=s
    s=compareModels(tmpdir,verbose=verbose)
    output_str=output_str+s
    equal=s.endswith('TRUE')

    counterExample=None
    outputsModel1=None
    outputsModel2=None


    if not equal:
        (inputs,outputs)=getTraceFromText_IOMEALY(s) 
        counterExample=inputs
        outputsModel2=outputs

        if bothways:
            # redo compare with models switched to get outputs of model1
            s=compareModels(tmpdir,True, verbose=verbose)
            output_str=output_str+s
            (inputs,outputs)=getTraceFromText_IOMEALY(s) 
            #counterExample=inputs
            outputsModel1=outputs

    if not keeptmp:
        shutil.rmtree(tmpdir)

    # remove the control chars from the cadp output
    output_str=''.join(c for c in output_str if ord(c) >= 10)

    return {'equal':equal,'inputsCounterExample':counterExample,'outputsModel1':outputsModel1,'outputsModel2':outputsModel2,'fullOutput':output_str}

def main(args):
    """

    """
    if '--test' in args: 
        test(args)
        return

    (modelfile1,modelfile2,range_min,range_max,keeptmp,verbose,skip_bothways,output_type,model_type)=args_parsing(args)

    bothways = not skip_bothways
    if model_type==ModelType.IOMEALY:
        result= compare_IOMEALY(modelfile1,modelfile2,range_min,range_max,keeptmp,verbose,bothways)
        if output_type==OutputType.BOOLEAN:
            print result['equal']
        elif output_type==OutputType.TEXT:
            print "equal: ", result['equal']
            if not result['equal']:
                print "counterexample:"
                print "  inputs: ", result['inputsCounterExample']
                print "  outputs model 1 :", result['outputsModel1']
                print "  outputs model 2 :", result['outputsModel2']
            if verbose:
                print ""
                print "verbose output:" 
                print result['fullOutput']
        else:
            import json
            print json.dumps(result, sort_keys=True, indent=4)
    else:
        result= compare_lts(modelfile1,modelfile2,range_min,range_max,keeptmp,verbose)
        if output_type==OutputType.BOOLEAN:
            print result['equal']
        elif output_type==OutputType.TEXT:
            print "equal: ", result['equal']
            if not result['equal']:
                print "counterexample:"
                print "   actions model 1 :", result['actionsModel1']
                print "   Note: last action not in model 2"
            if verbose:
                print ""
                print "verbose output:" 
                print result['fullOutput']
        else:
            import json
            print json.dumps(result, sort_keys=True, indent=4)

def test(args):
    print "test"
    
    print"compare lts ;  compare --lts iomealy.a.dot iomealy.b.dot)"
    main(['compare','--lts','--output-text','testinput/iomealy.a.dot','testinput/iomealy.b.dot'])
    print"compare iomealy in lts ;  compare --iomealy iomealy.a.dot iomealy.b.dot)"
    main(['compare','--output-text','testinput/iomealy.a.dot','testinput/iomealy.b.dot'])

    print"compare register (equal)"
    main(['compare','--output-text','testinput/register.orig.xml','testinput/register.learned.xml'])
    print"compare register (inequal)"
    main(['compare','--output-text','testinput/register.orig.xml','testinput/register.mutated.xml'])


if __name__ == "__main__":
    try:
        test(sys.argv)
#         if '--test' in sys.argv: 
#             test(sys.argv)
#         else:    
#             main(sys.argv)
    except KeyboardInterrupt as e:
        print >> sys.stderr, "\n\nExiting on user cancel."
        sys.exit(1)

