'''
 ioaut2mealydot 

'''

import re
import argparse
     
 
# load libraries from ./lib
import lib.util,lib.check,lib.run,lib.error

class Transition(object):
    def __init__(self, source, target):
        self.source=source
        self.target=target
       
    def __str__(self):
       return str(self.source) + " -- " +  self.getLabel() + " --> " + str(self.target)
              
    def to_dot(self):
        dotline="  %s -> %s [label=\"%s\"];" % (self.source,self.target,self.getLabel())
        return dotline

    def getLabel(self):
        pass
        
class MealyTransition(Transition):
   """
   Mealy transition consisting of :
      - source : start state of transition
      - target : end state of transition
      - input  : input of transition
      - output  : output of transition
   """
   def __init__(self, source, target, input,output):
       self.source=source
       self.target=target
       self.input=input
       self.output=output

   def __repr__(self):
       return "MealyTransition(" + repr(self.source) + "," + repr(self.target) + "," + repr(self.input)  + "," + repr(self.output) + ")"


   def getLabel(self):
       return  self.input + "/" + self.output

class IOTransition(Transition):
      pass

class InputTransition(IOTransition):
   """
   Input transition consisting of :
      - source : start state of transition
      - target : end state of transition
      - input  : input of transition
   """
   def __init__(self, source, target, input):
       self.source=source
       self.target=target
       self.input=input

   def __repr__(self):
       return "InputTransition(" + repr(self.source) + "," + repr(self.target) + "," + repr(self.input)  + ")"
       


   def getLabel(self):
       return  self.input 
       
class OutputTransition(IOTransition):
   """
   Output transition consisting of :
      - source : start state of transition
      - target : end state of transition
      - output  : output of transition
   """
   def __init__(self, source, target,output):
       self.source=source
       self.target=target
       self.output=output

   def __repr__(self):
       return "InputTransition(" + repr(self.source) + "," + repr(self.target) + ","  + repr(self.output) + ")"
       
   def __str__(self):
       return str(self.source) + " -- " +  self.getLabel() + " --> " + str(self.target)

   def getLabel(self):
       return   self.output   
      
class State:
    def __init__(self, name):
        self.name = name

    def getId(self):
        return self.name

    def to_dot(self, initial=False ):
        if initial == True:
            return "  %s [peripheries=2,color=\"red\"];" % (self.name)
        else:
            return "  %s;" % (self.name)
      
class Automata(object):
   """
   Automata
      has only startState, states and transitions
      Note: most support serialization of transition and states
   """
   def __init__(self, startState=None,transitions=[]):
       self.startState=startState
       self.transitions=transitions
       
       locations=set() 
       for trans in self.transitions:
             locations.add(trans.source)
             locations.add(trans.target)
       self.states=[] 
       for location in locations:
           self.states.append(State(location))             

   def __str__(self):
       """string representation of Automata
       
       prints start state and then all transitions between states
       """
       strings=["startState=" + str(self.startState)]
       for t in self.transitions:
           strings.append(str(t))
       return "\n".join(strings)
       
   def to_dot(self):
        locations="\n".join([s.to_dot(initial=(s.getId()==self.startState)) for s in self.states])
        transitions="\n".join([t.to_dot() for t in self.transitions])
        dotstr="digraph g {\nnode [shape=oval];\n%s\n%s\n}\n" % (locations,transitions)
        return dotstr

   def to_dotfile(self,filename):
       f=open(filename,'w')
       f.write(self.to_dot())
       f.close()
   

class MealyFSM(Automata):
   """
   Finite state machine
      has only startState, states and transitions
      Note: supports input/output label on transitions
   """
   pass
       
class IOFSM(Automata):
   """
   Finite state machine
      has only startState, states and transitions
      Note: supports input and output transitions
   """
   def __init__(self, startState=None,inputtransitions=[],outputtransitions=[]):
       self.startState=startState
       self.inputtransitions=inputtransitions      
       self.outputtransitions=outputtransitions  
       self.transitions= self.inputtransitions + self.outputtransitions
 
   def getMealyFSM(self):
       """ midstate is state after input and before an output
           for a deterministic automata each midstate should only have 
           one outgoing output transition!
       """
       midstate2outputtrans={}
       for outputtrans in self.outputtransitions:
           midstate2outputtrans[outputtrans.source]=outputtrans

       newtrans=[]
       for inputtrans in self.inputtransitions:
           outputtrans=midstate2outputtrans[inputtrans.target]    
           newtrans.append(MealyTransition(inputtrans.source, outputtrans.target, inputtrans.input,outputtrans.output)) 

       return MealyFSM(self.startState,newtrans)
 
# can also be used for tomte_aut2cadp_aut
def get_ioautomata_from_autfile(infile):
    """  get io automata from aut file 

      .aut format is aldebaran format 
       see http://www.mcrl2.org/release/user_manual/language_reference/lts.html#aldebaran-format
       
    """
    
    f=file(infile)
    lines=f.readlines()

    # first line: 'des (first_state,nr_of_transitions,nr_of_states)'
    firstline=lines[0].strip()
    regexpr_firstline=re.compile(r'\s*des\s*\(\s*(\d*)\s*,\s*(\d*)\s*,\s*(\d*)\s*\)')
    match=regexpr_firstline.match(firstline)
    #print firstline
    if match:
        match=match.groups()
        start_state=match[0]
        nr_of_transitions=match[1]
        nr_of_states=match[2]
        #print "start_state",start_state
        #print "nr_of_transitions",nr_of_transitions
        #print "nr_of_states",nr_of_states
    else:
        print "error: problem reading aut file"
        import sys
        sys.exit(1)
       
    transitions=lines[1:]
    regexpr_transition=re.compile(r'\s*\(\s*(\d*)\s*,\s*([^,]*)\s*,\s*(\d*)\s*\)')
    inputtransitions=[]
    outputtransitions=[]
    for transition in transitions:
        match=regexpr_transition.match(transition)
        if match:
            match=match.groups()
            #label=regexpr_tag.sub('',match[2])
            label=match[1].strip('"').replace(' !+','_').replace(' !','_')


            source=match[0]
            target=match[2]
            # assume inputs start with i
            # assume outputs start with o
            if label[0].lower()=='i':
                 input="I"+label[1:]
                 inputtrans=InputTransition(source,target,input)
                 inputtransitions.append(inputtrans)
            else:
                 output="O"+label[1:]
                 outputtrans=OutputTransition(source,target,output)
                 outputtransitions.append(outputtrans)
            
    return IOFSM(start_state,inputtransitions,outputtransitions)

     
def get_mealy_from_dotfile_from_learnlib(dot_file):
    """ Get mealy transition system from graphviz dot file

        The dot file:
           - describes a mealy machine as a digraph with labels
           - label is of form :  <input>/<ouput>
           - encodes the start state with the color='red' attribute
             note: this corresponds with the highlighted state in learnlib API


        Returns: [start_state,transions]
        Where :
           - start_state: start state label
           - transitions: list of transitions

    """

    start_state='unknown'
    f=file(dot_file)
    lines=f.readlines()

    # find start state
    for line in lines:
        line=line.replace('label=""','')
        if line.find('[color="red"]') != -1:
            start_state=line[:line.find(' ')]
            break


    # get transitions
    transitions=[]
    for line in lines:
        if line.find('->') != -1:
            transitions.append(line)

    # throw away transitions with the keywords : quiescence or inconsistency or undefined
    #transitions = [ t for t in transitions if ( 'quiescence' not in t ) and ( 'inconsistency' not in t  )  and ( 'undefined' not in t  )]

    trans_out=[]
    regexpr_transition=re.compile(r'\s*(\w*)\s*-\>\s*(\w*)\s*\[label=\<(.*)\>\]')
    regexpr_tag=re.compile(r'<[^>]+>')
    for transition in transitions:
        match=regexpr_transition.match(transition)
        if match:
            match=match.groups()
            label=regexpr_tag.sub('',match[2])
            trans_out.append( LabeledTransition(
              source = match[0],
              target = match[1],
              label = label
            ))


    states=set()
    for t in trans_out:
        states.add(t.source)
        states.add(t.target)


    return LTS(start_state,states,trans_out)
       

       
def args_parsing(args):
    import os.path
    argparser = argparse.ArgumentParser(prog=os.path.basename(args[0]))

    argparser.add_argument('autmodel', help='aut model file')
    argparser.add_argument('dotmodel', help='dot model file')
           
    parse_result=argparser.parse_args(args[1:])

    # handle input/output file argument
    modelfile_aut=os.path.abspath(parse_result.autmodel).replace('\\','/')
    modelfile_dot=os.path.abspath(parse_result.dotmodel).replace('\\','/')  
    
    lib.check.check_isreadablefile(modelfile_aut)
    return  [modelfile_aut,modelfile_dot]

def main(args):
    """

    """

    (modelfile_aut,modelfile_dot)=args_parsing(args)

    iomodel=get_ioautomata_from_autfile(modelfile_aut)
    mealymodel=iomodel.getMealyFSM()
    mealymodel.to_dotfile(modelfile_dot) 




if __name__ == "__main__":
    try:
        main(sys.argv)
    except KeyboardInterrupt as e:
        print >> sys.stderr, "\n\nExiting on user cancel."
        sys.exit(1)
               