#!/usr/bin/env python
"""
generates aut for jtorx from dot file learned with learnlib
  '?COIN_1_1'
  '!TEA_0_1'

note: dot file uses I for input instead of ? and O for output instead of !
"""
# Author: Harco Kuppens


import sys, os, re, pprint  # modules from standard lib (python 2.6 and later)
import argparse
import util
import re


def get_start_state(lines):
    """
      find start state
      
      return:
           start state found or "unknown" if start state not found
      
    """   
    
    # try:  (tomte with old learnlib)
    #  start state colored red!  
    for line in lines:
        line=line.replace('label=""','')
        if line.find('color="red"') != -1:
            start_state=line[:line.find('[')].strip()
            return start_state
        
    # try: (new learnlib)          
    #  line in dot: __start0 -> s0;
    for line in lines:
        if line.find('->') != -1:
            if line.find('__start') != -1:
                start_state=line[line.find('->')+2:].strip(" ;\t\n")
                return start_state 
                   
    start_state='unknown'
    return start_state

def get_lts_from_dotfile(dot_file):
    """ Get labeled transition system from graphviz dot file

        The dot file:
           - describes a digraph with labels
           - encodes the start state with the color='red' attribute
             note: this corresponds with the highlighted state in learnlib API

        Returns: [start_state,transitions]
        Where :
           - start_state: start state label
           - transitions: list of transitions
                 where each transition is a dictionary
                 with the fields :
                      'source' : source state of transition
                      'target' : target state of transition
                      'label': label of transition


    examples:
    
    
    
     old learnlib
     
         digraph G {
         label=""
         s0 [color="red"]
         s1
         s2
         s3
         s4
         s0 [label="s0"];
         s0 -> s0[label=<<table border="0" cellpadding="1" cellspacing="0"><tr><td>IBUTTONMARS</td><td>/</td><td>Oquiescence</td></tr></table>>]
         s0 -> s0[label=<<table border="0" cellpadding="1" cellspacing="0"><tr><td>IBUTTONSNICKERS</td><td>/</td><td>Oquiescence</td></tr></table>>]
         s0 -> s1[label=<<table border="0" cellpadding="1" cellspacing="0"><tr><td>ICOIN10</td><td>/</td><td>OOK</td></tr></table>>]
         s0 -> s2[label=<<table border="0" cellpadding="1" cellspacing="0"><tr><td>ICOIN5</td><td>/</td><td>OOK</td></tr></table>>]
         s0 -> s0[label=<<table border="0" cellpadding="1" cellspacing="0"><tr><td>IREFUND</td><td>/</td><td>Oquiescence</td></tr></table>>]
         s1 [label="s1"];
         s1 -> s1[label=<<table border="0" cellpadding="1" cellspacing="0"><tr><td>IBUTTONMARS</td><td>/</td><td>Oquiescence</td></tr></table>>]
         s1 -> s1[label=<<table border="0" cellpadding="1" cellspacing="0"><tr><td>IBUTTONSNICKERS</td><td>/</td><td>Oquiescence</td></tr></table>>]
         ...
         s4 -> s3[label=<<table border="0" cellpadding="1" cellspacing="0"><tr><td>ICOIN5</td><td>/</td><td>OOK</td></tr></table>>]
         s4 -> s0[label=<<table border="0" cellpadding="1" cellspacing="0"><tr><td>IREFUND</td><td>/</td><td>OMONEY</td></tr></table>>]
         }
     
     
     new learnlib
     
         digraph g {
         __start0 [label="" shape="none"];
    
             s0 [label="0" shape="circle"];
            ...
             s7 [label="7" shape="circle"];
             s0 -> s1 [label="SIGNON 8:01 / 251"];
             s0 -> s0 [label="SIGNOFF / 550"];
            ...
             s7 -> s7 [label="TRANS TM_BANK / 550"];
             s7 -> s7 [label="IDLE / 531"];
    
         __start0 -> s0;
         } 


    """

    
    f=file(dot_file)
    lines=f.readlines()

    start_state=get_start_state(lines)
    if start_state=='unknown':
        start_state='s0'
        print "---->  GUESSED STARTSTATE : could be wrong"

    # get transitions
    transitions=[]
    for line in lines:
        if line.find('__start') != -1:
            continue # skip transition to start state in newlearnlib dot
        if line.find('->') != -1:
            transitions.append(line)

    # throw away transitions with the keywords : quiescence or inconsistency or undefined
    #transitions = [ t for t in transitions if ( 'quiescence' not in t ) and ( 'inconsistency' not in t  )  and ( 'undefined' not in t  )]

    trans_out=[]
    regexpr_transition=re.compile(r'\s*(\w*)\s*-\>\s*(\w*)\s*\[\s*label\s*=\s*[\<"](.*)[\>"]\s*,?\s*.*\]')
    regexpr_tag=re.compile(r'<[^>]+>')
    for transition in transitions:
        #print "transition", transition
        match=regexpr_transition.match(transition)
        if match:
            match=match.groups()
            label=regexpr_tag.sub('',match[2])
            trans_out.append({
              'source' : match[0].strip(),
              'target' : match[1].strip(),
              'label': label.strip()
            })

    states=set()
    for t in trans_out:
        states.add(t['source'])
        states.add(t['target'])


    return [start_state,states,trans_out]


def parse_labels_of_mealy_lts(transitions):
    """Parse labels of labeled transition system

    Just splits label into input and output label,
    but doesn't split the transition.

    from

       transition = {
           'source' : <source state label>
           'target' : <target state label>
           'label'  : <mealy transition label>
       }
  
    to

       transition = {
           'source' : <source state label>
           'target' : <target state label>
           'input'  : <input label>
           'output' : <output label>
       }

    """
    trans_out=[]
    for t in transitions:
        label=t['label']
        [inputstr,outputstr]=label.split('/')
        trans_out.append({
          'source' : t['source'].strip(),
          'target' : t['target'].strip(),
          'input':  inputstr.strip(),
          'output':  outputstr.strip(),
        })
    return trans_out


def mark_input_output_labels_of_mealy_lts_with_question_and_exclamation_mark(transitions):
    """ Add mark in input and output labels to easily see if a label is input or output :
        
        for 
           input:   xxx -> ?xxx       prefix input with ?
           output:  yyy -> !yyy       prefix output with !
        
    
    """
    trans_out=[]
    for t in transitions:
        trans_out.append({
          'source' : t['source'],
          'target' : t['target'],
          'input':  "?" + t['input'],
          'output': "!" + t['output'],
        })
    return trans_out

def mark_input_output_labels_of_mealy_lts_with_I_and_O_mark(transitions):
    """ Add mark in input and output labels to easily see if a label is input or output :
        
        for 
           input:   xxx -> Ixxx       prefix input with I
           output:  yyy -> Oyyy       prefix output with O
        
    
    """
    trans_out=[]
    for t in transitions:
        trans_out.append({
          'source' : t['source'],
          'target' : t['target'],
          'input':  "?" + t['input'],
          'output': "!" + t['output'],
        })
    return trans_out

# not used
def mark_input_output_labels_of_mealy_lts_with_question_and_exclamation_mark_when_IO_marked(transitions):
    """ Change marking of input and output labels as follows :
        
        for 
         input  : Ixx  ->  ?xx
         output : Oyy  ->  !yy  
    
    """
    trans_out=[]
    for t in transitions:
        trans_out.append({
          'source' : t['source'],
          'target' : t['target'],
          'input':  "?" + t['input'][1:],
          'output':  "!" + t['output'][1:],
        })
    return trans_out


def split_io_transitions_in_separate_input_and_output_transition(io_transitions,nr_states):
    """Split transitions with both an input and output event into two transitions

       Makes two sequential transitions with a dummy state in between:
        - dummy state <midstate> is labeled :
             m_<counter>
        - first transition :
             <source> -> <midstate>  for  <input>
        - second transition :
             <midstate> -> <target>  for <output>
    """
    trans_out=[]
    id=nr_states
    for t in io_transitions:
        midstate= 'm' + str(id)
        trans_out.append({
            'source': t['source'].strip(),
            'target': midstate.strip(),
            'label' : t['input'].strip(),
        })
        trans_out.append({
            'source': midstate.strip(),
            'target': t['target'].strip(),
            'label' : t['output'].strip(),
        })
        id=id+1

    states=set()
    for t in trans_out:
        states.add(t['source'])
        states.add(t['target'])

    return [states,trans_out]


def transitions2aut(transitions,first_state,states):
    nr_of_states=len(states)
    nr_of_transitions=len(transitions)
   
    # make list of states where the start state is the first state
    states=list(states)
    states.remove(first_state)
    states= [first_state]+ states
    # give each state string an unique index 
    indexes=[str(i) for i in range(len(states)) ]
    state2index = dict( zip( states, indexes ))

    strings=[ "des(" + state2index[first_state] + ","  + str(nr_of_transitions) + "," + str(nr_of_states) +")"]
    for t in transitions:
        #aut_edge ::= "(" start_state "," label "," end_state ")"
        strings.append("("+state2index[t['source']]+ "," + '"' + t['label'] + '"' + "," + state2index[t['target']] + ")" )

    return "\n".join(strings)

def args_parsing(args):

    argparser = argparse.ArgumentParser(prog=os.path.basename(args[0]))

    argparser.add_argument('dotfile', help='dot file learned with learnlib')
    argparser.add_argument('autfile', help='aut output file (aldebaran format)')

    str=argparser.format_help()

    def myhelp( ):
        print str

    def myerror( message):
            print >>sys.stderr, ('%s: error: %s\n') % (argparser.prog, message)
            argparser.print_usage(sys.stderr)
            exit(1)

    argparser.error=myerror
    argparser.format_help=myhelp

    parse_result=argparser.parse_args(args[1:])

    # handle input/output file argument
    dot_file=os.path.abspath(parse_result.dotfile).replace('\\','/')
    aut_file=os.path.abspath(parse_result.autfile).replace('\\','/')
    # Note: on windows also interprets / as dir sep, and returns path with windows \

    util.check_file_extension(dot_file,'dot')
    util.check_file_extension(aut_file,'aut')

    return  [dot_file,aut_file]

def lts_dot2aut(infile,outfile):
    """
       convert a LTS automata in dot format to aut format
    """
    [start_state,states,transitions]=get_lts_from_dotfile(infile)
    result=transitions2aut(transitions,start_state,states)

    f=open(outfile ,'w')
    f.write(result)
    f.close()


def dot2jtorxaut(infile,outfile):
    """
       convert an IO mealy machine in dot-LTS format which has a single
       transition for both input and output with 
       
          label format : "input/output"
          
       to and IO mealy machine in aut-LTS format which has input
       and output on separation transitions ( and thus labels).
       
       Thus:  
            StartState --  "input/output" --> EndState
       converted to     
            StartState --  "?input" --> MidState -- "!output" --> EndState
            
       note: for jtorx  ? and ! markers are added to the labels
             to mark input and output explicitlys     
       
    """
    [start_state,states,transitions]=get_lts_from_dotfile(infile)
    io_transitions=parse_labels_of_mealy_lts(transitions) # each transition has input and output
    io_transitions=mark_input_output_labels_of_mealy_lts_with_question_and_exclamation_mark(io_transitions)
    [states,transitions]=split_io_transitions_in_separate_input_and_output_transition(io_transitions,len(states)) # each transition only has label again

    result=transitions2aut(transitions,start_state,states)

    f=open(outfile ,'w')
    f.write(result)
    f.close()


def main(args):
    """
       from mealy machine in a .dot file written by DotUtil.write of learnlib
       we create an .aut file containing an lts where input and output each
       have its own labeled transition. An input transition has a
       label starting with '?' and an output transition has a label
       starting with '!'

    """

    (dot_file,aut_file)=args_parsing(args)

    dot2jtorxaut(dot_file,aut_file)



if __name__ == "__main__":
    #main(sys.argv)
    print "CandyMachine.dot"
    main(["sut__learnlib_dot2jtorx_aut", "CandyMachine.dot","CandyMachine.aut"])
    print "ScannerGraph.dot"
    main(["sut__learnlib_dot2jtorx_aut","ScannerGraph.dot","ScannerGraph.aut"])
    print "server.dot"
    main(["sut__learnlib_dot2jtorx_aut","server.dot","server.aut"])
    print "client.dot"
    main(["sut__learnlib_dot2jtorx_aut","client.dot","client.aut"])

