#!/usr/bin/env python
"""
  generate jar file from a model file which implements that model
  using a socket interface
"""
# Author: Harco Kuppens


import sys, re, pprint,os  # modules from standard lib (python 2.6 and later)

import tempfile
import shutil
import yaml
import argparse

import util
import model_parser
import gentemplate




def check_dependencies():
    util.check_program('java','.exe')
    util.check_program('javac','.exe')

def args_parsing(args):

    argparser = argparse.ArgumentParser(prog=os.path.basename(args[0]))

    argparser.add_argument("--source", nargs='?' , metavar='dir', default=False, const='TEMPDIR', help="only generate the source for building the jar file; no jar file will be build")
    #argparser.add_argument("--source-only", action='store_true',help="only generate the source for building the jar file")
    #argparser.add_argument("--source-dir", action='store',help="output the generated source for building the jar file in the give source directory")

    argparser.add_argument('modelfile', help='input uppaal model file')
    argparser.add_argument('jarfile', nargs='?', default='out.jar', help='output jar file implementing model')
    argparser.add_argument('package', nargs='?', default='mypackage', help='java package of java files implementing model; we can then easily generate different suts in different namespaces such that they do not conflict each other')
    argparser.add_argument('className', nargs='?', default='myclass',  help='java class name of java file implementing the SutInterface java interface; as for package we can distinguish different implementations by class name')


    parse_result=argparser.parse_args(args[1:])


    # handle input/output file argument
    modelfile=os.path.abspath(parse_result.modelfile).replace('\\','/')
    jarfile=os.path.abspath(parse_result.jarfile).replace('\\','/')
    # Note: on windows also interprets / as dir sep, and returns path with windows \
    package=parse_result.package
    className=parse_result.className
    source_dir=parse_result.source


    util.check_isreadablefile(modelfile)
    
    source_only=True
    if not source_dir: 
       source_only=False
       source_dir='TEMPDIR'

    if source_dir== 'TEMPDIR': 
         source_dir=tempfile.mkdtemp(prefix="sut_uppaal2jar_") # creates source dir directly
    else: 
         source_dir=os.path.abspath(source_dir).replace('\\','/')
         util.check_noneexistingdir(source_dir)
         os.mkdir(source_dir)

    return  [modelfile,jarfile,package,className,source_dir,source_only]

def add_to_dict_with_list_value(dict,key,value):
    if not dict.has_key(key):
      dict[key]=[]
    dict[key].append(value)

def combineGuards(guard1,guard2):
    if not guard1 or not guard2: 
       return guard1+guard2
    else:
       return "( {0} ) && ( {1} )".format(guard1,guard2)    
    
    

 
def uppaal2java(srcdir,modelfile,package,className):
    """ 
      generate a jar file from a model file which implements that model
      using a socket interface
    """
    sep=os.path.sep


    # 1. generate java source from model
    #------------------------------------
    
    # template dir is relative from this module :
    template_dir = os.path.dirname(os.path.abspath(__file__))   + '/templates'

    template_files= [
       "Main.java",
       "Sut.java",
       "Parameter.java",
       "InputAction.java",
       "OutputAction.java",
       "Action.java",
    ]
        

    # date from model
    parsed_data=model_parser.parse_model(modelfile)
    
    
    
    #    |   input       note: single input can be shared by  multiple output actions (each their own output guard!) 
    #   /|\  outputs!       
    #
    # 
    #  \|/  inputs    note: single output can be share by multiple input actions!
    #   |   output
    #
    #  => for each midpoint :
    #      1. get all input trans leading to it, and get all output trans leaving from it
    #      2. take direct product of these => set of mealy transitions (no midstate anymore!) 

    inputWithSourceState2inputTransitions=parsed_data['input_with_location2guard_with_action_with_target'];
    midstate2outputTransitions=parsed_data['location2guard_with_action_with_output_with_target'];
    # note: both an input as an output transaction have guards and updates. 
    #       though an output transaction can only have the input parameters in update through temporary helper variables
    
    midstate2leadingTransitions={}
    locations=set()
    for inputMethod,sourceState2inputTransitions in inputWithSourceState2inputTransitions.items(): 
        for sourceState,inputTransitions in  sourceState2inputTransitions.items():
            locations.add(sourceState)
            for inputTransition in inputTransitions:
                add_to_dict_with_list_value(midstate2leadingTransitions,inputTransition['target'],inputTransition)
                
    parsed_data['locations']=locations
                
         

    state2input2inputtransitions={}  # note: input transition has a set of output transitions which guard can only be tester after the inputtransition's update has been executed 
    for location in locations:
        state2input2inputtransitions[location]={}
        #for inputmethod in inputInterfaces:
       
    for midstate,inputTransitions in midstate2leadingTransitions.items():
        outputTransitions=midstate2outputTransitions[midstate]
        for inputTransition in inputTransitions:
            inputTransition['outputtransitions']=outputTransitions  # note: input transition has a set of output transitions which guard can only be tester after the inputtransition's update has been executed 
            stateID=inputTransition['source']
            input2inputtransitions = state2input2inputtransitions[stateID]
            add_to_dict_with_list_value(input2inputtransitions,inputTransition['name'],inputTransition)
    
        
    
    
    del(parsed_data['input_with_location2guard_with_action_with_target'])
    del(parsed_data['location2guard_with_action_with_output_with_target'])
    parsed_data['state2input2inputtransitions']=state2input2inputtransitions
    parsed_data['model_file']=modelfile.replace('\\','\\\\')
    parsed_data['package']=package
    parsed_data['className']=className


    # make all dirs for all packages
    packagedirs=package.split(".")
    pdir=srcdir 
    for subdir in packagedirs:
       pdir=pdir + sep + subdir
       os.mkdir(pdir)

    # fill template files
    for template_file in template_files:
        if template_file == "Sut.java":
            output_file = pdir + sep + className + ".java"
        else:
            output_file = pdir + sep + template_file 
        gentemplate.fill_template(output_file,parsed_data,template_dir,template_file)


def uppaal2jar(modelfile,jarfile,package,className,outdir,source_only):
    """ 
      generate a jar file from a model file which implements that model
      using a socket interface
    """
    sep=os.path.sep

        

    # 1. generate from model the java source 
    #---------------------------------------
    # src directory to generate java source in
    gen_srcdir=outdir + sep + 'src' + sep
    os.mkdir(gen_srcdir)

    uppaal2java(gen_srcdir,modelfile,package,className)

    # 2. generate manifest file
    #--------------------------
    # manifest file for jar
    jar_manifestfilename =  'manifest.txt'
    jar_manifestfile =  outdir  + sep + jar_manifestfilename
    
    # template dir is relative from this module :
    template_dir = os.path.dirname(os.path.abspath(__file__))   + '/templates'
    jar_manifest_template = 'manifest.txt' 
    parsed_data={ 'package' : package }
    gentemplate.fill_template(jar_manifestfile,parsed_data,template_dir,jar_manifest_template)


    if source_only: return 

    # 3. compile all java source
    #----------------------------

    jar_manifestfile =  outdir  + sep + jar_manifestfilename

    """
      example of java compilation  :
            mkdir bin
            javac -sourcepath src/:sutsrc/ -cp lib/learnlib-distribution-20110714.jar -d bin  src/sut/Main.java
            jar -cvfm sut.jar manifest.txt -C bin  /

            Where dirs
              - src : source files in sut resource directory
              - sutsrc : tempdir where source of generated sut/Sut.java is located
              - bin : tempdir where binaries are generate
    """

    # resource dir is relative from this module :
    resource_dir = os.path.abspath( os.path.dirname(os.path.abspath(__file__)) + '/../resources/' )
    # second src directory which contains also code  which is also used to build jar file
    static_srcdir= resource_dir  + sep +  'sutjar' +  sep + 'src'

    # java dir (containing jar library files) is relative from this module :
    java_dir = os.path.abspath( os.path.dirname(os.path.abspath(__file__)) + '/../java/' )


    # java source path (classpath= comma separated list of srcdirs)  :
    java_srcpath=static_srcdir + os.path.pathsep + gen_srcdir

    # java bin directory to put compiled class files in
    java_bindir=outdir  + sep + 'bin'  + sep
    os.mkdir(java_bindir)

    # java main file
    java_mainFile = gen_srcdir + package.replace('.',sep) + sep + 'Main.java'

    loggerjar= java_dir  + sep +  'log4j-1.2.17.jar'

    # use  run_command_with_output and discard the returned output,
    # so any normal output won't be printed to the console
    # note: use cwd='/' to be safe
    util.run_command_with_output([ 'javac','-source', '1.7','-target', '1.7', '-sourcepath' , java_srcpath , '-cp', loggerjar , '-d', java_bindir  , java_mainFile  ],cwd='/')
    util.run_command_with_output([ 'jar', '-cvfm', jarfile , jar_manifestfile , '-C', java_bindir ,  sep ], cwd='/')





def main(args):
    """
      generate jar file from a model file which implements that model
      using a socket interface
    """
    check_dependencies()

    (modelfile,jarfile,package,className,source_dir,source_only)=args_parsing(args)
    
    
    uppaal2jar(modelfile,jarfile,package,className,source_dir,source_only)


    if source_only: 
        print "created sources in directory: " +  source_dir
        print "no jar file created"
    else:
        print "created jar file: " +  jarfile
        shutil.rmtree(source_dir)

if __name__ == "__main__":
    print main(sys.argv)

